from typing import Dict, Any
from typing import Protocol

LambdaDict = Dict[str, Any]


class LambdaCognitoIdentity(Protocol):
    cognito_identity_id: str
    cognito_identity_pool_id: str


class LambdaClientContextMobileClient(Protocol):
    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


class LambdaClientContext(Protocol):
    client: LambdaClientContextMobileClient
    custom: LambdaDict
    env: LambdaDict


class LambdaContext(Protocol):
    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: LambdaCognitoIdentity
    client_context: LambdaClientContext

    @staticmethod
    def get_remaining_time_in_millis() -> int:
        """protocol method"""



from dataclasses import dataclass, field
from typing import Protocol, List


@dataclass()
class EmailAttachment:
    content: str
    name: str


@dataclass
class Email:
    sender: str
    recipients: List[str]
    html_body: str
    raw_text_body: str
    subject: str
    attachments: List[EmailAttachment] = field(default_factory=list)


class Mailer(Protocol):
    async def send(self, email: Email) -> None:
        """protocol function"""



from typing import TypeVar, Callable, Awaitable, Any, Optional

from pydantic import BaseSettings

from clubbi_utils.common.async_messenger import AsyncSenderStr
from slack_sdk.web.async_client import AsyncWebClient

from clubbi_utils.logger import logger

RT = TypeVar("RT")


class SlackMessenger(AsyncSenderStr):
    def __init__(self, token: str, channel_name: str):
        self._client = AsyncWebClient(token=token)
        self._channel_name = channel_name

    async def send(self, message: str) -> None:
        await self._client.chat_postMessage(
            channel=self._channel_name,
            text=message,
        )

    async def send_message_as_file(self, content: str, file_name: Optional[str], title: Optional[str],
                                   message: Optional[str]) -> None:
        await self._client.files_upload(channels=self._channel_name,
                                        content=content,
                                        filename=file_name,
                                        title=title,
                                        initial_comment=message)


class SlackSettings(BaseSettings):
    environment: str
    slack_channel_name: str = ""
    slack_token: str = ""


class LocalLoggerMessager(AsyncSenderStr):
    @staticmethod
    async def send(message: str) -> None:
        logger.info(f"LoggerMessager: {message}")


def with_slack_messenger(f: Callable[..., Awaitable[RT]]) -> Callable[..., Awaitable[RT]]:
    async def wrapper(*args: Any, **kwargs: Any) -> RT:
        settings = SlackSettings()
        slack_messenger: AsyncSenderStr

        if settings.environment == "local":
            slack_messenger = LocalLoggerMessager()
        else:
            assert settings.slack_token and settings.slack_channel_name
            slack_messenger = SlackMessenger(settings.slack_token, settings.slack_channel_name)
        return await f(slack_messenger, *args, **kwargs)

    return wrapper



# Docs: https://docs.aws.amazon.com/lambda/latest/dg/with-s3-example.html#with-s3-example-test-dummy-event
from typing import TypedDict, List


class AwsS3EventBucket(TypedDict):
    name: str
    arn: str


class AwsS3EventObject(TypedDict):
    key: str
    size: int
    eTag: str
    sequencer: str


class AwsS3EventInfo(TypedDict):
    s3SchemaVersion: str
    configurationId: str
    bucket: AwsS3EventBucket
    object: AwsS3EventObject


class AwsS3EventRecord(TypedDict):
    eventVersion: str
    eventSource: str
    awsRegion: str
    eventTime: str
    eventName: str
    s3: AwsS3EventInfo


class AwsS3Event(TypedDict):
    Records: List[AwsS3EventRecord]



from typing import Optional, Type, TypeVar, Any

from pydantic import parse_obj_as, ValidationError

from clubbi_utils.common.clubbi_validation_error import ClubbiValidationError, ErrorUnit

T = TypeVar("T")


def parse_obj_list_with_errors(type_: Type[T], obj: Any) -> tuple[list[T], Optional[ClubbiValidationError]]:
    """Parses a list of `obj` given a type `T` and returns a tuple containing the parsed result and errors,
    if a parsing error occurs the current value will be skiped and it's error will be appended in the errors result.

    Example:
    ```python
    from pydantic import BaseModel
    from clubbi_utils.common.parse_obj_list_with_errors import parse_obj_list_with_errors
    from clubbi_utils.common.clubbi_validation_error import ClubbiValidationError


    class Foo(BaseModel):
        a: int


    data = [{"a": 1}, {"a": "not a number"}, {"a": 2}]
    parsed_values, error = parse_obj_list_with_errors(Foo, data)

    assert parsed_values == [Foo(a=1), Foo(a=2)]
    assert error ==  ClubbiValidationError(
        errors=[
            {
                "identifier": "linha 2",
                "msg": "value is not a valid integer",
                "loc": "a",
                "type": "type_error.integer",
            }
        ],
        model="Foo",
    )
    ```
    """

    parsed_values: list[T] = []
    errors: list[ErrorUnit] = []
    type_qualname = type_.__qualname__
    if not isinstance(obj, list):
        validation_error = ClubbiValidationError(
            errors=[
                ErrorUnit(
                    identifier="0",
                    msg="Obj is not a list",
                    loc="0",
                    type="ValidationError",
                )
            ],
            model=type_qualname,
        )
        return [], validation_error
    for row_number, datum in enumerate(obj):
        try:
            parsed_value = parse_obj_as(type_, datum)
            parsed_values.append(parsed_value)
        except ValidationError as e:
            errors.extend(
                [
                    ErrorUnit(
                        identifier=f"linha {row_number + 1}",
                        msg=error["msg"],
                        loc=".".join(str(loc) for loc in error["loc"][1:]),
                        type=error["type"],
                    )
                    for error in e.errors()
                ]
            )
    if errors:
        return parsed_values, ClubbiValidationError(errors=errors, model=type_qualname)
    return parsed_values, None



from asyncio import StreamReader
from datetime import datetime
from typing import Union, AsyncIterator, List, Any, Dict

from aiobotocore.client import AioBaseClient
from aiobotocore.response import StreamingBody
from pydantic import BaseModel

_MIN_PART_SIZE = 5 * (1 << 20)


class ObjectAttributes(BaseModel):
    last_modified: datetime
    object_size: int


class S3ObjectStorage:
    def __init__(self, client: AioBaseClient, bucket: str, key_prefix: str = ""):
        self._client = client
        self._bucket = bucket
        self.key_prefix = key_prefix

    async def _put_object(self, name: str, data: bytes) -> str:
        key = self._build_key(name)
        await self._client.put_object(Bucket=self._bucket, Key=key, Body=data)
        return key

    def _build_key(self, name: str) -> str:
        return f"{self.key_prefix}{name}"

    async def _put_large_file(self, name: str, data: AsyncIterator[bytes]) -> str:
        key = self._build_key(name)
        mpu = await self._client.create_multipart_upload(Bucket=self._bucket, Key=key)
        upload_id = mpu["UploadId"]
        parts: List[Dict[str, Any]] = []

        async for chunk in _group_chunks_to_min_part_size(data):
            if chunk == b"":
                break
            part = await self._client.upload_part(
                Bucket=self._bucket, Key=key, PartNumber=len(parts) + 1, UploadId=upload_id, Body=chunk
            )
            parts.append({"PartNumber": len(parts) + 1, "ETag": part["ETag"]})

        await self._client.complete_multipart_upload(
            Bucket=self._bucket, Key=key, UploadId=upload_id, MultipartUpload=dict(Parts=parts)
        )
        return key

    async def put_object(self, name: str, data: Union[bytes, AsyncIterator[bytes]]) -> str:
        if isinstance(data, bytes):
            return await self._put_object(name, data)
        else:
            return await self._put_large_file(name, data)

    async def stream_object(
        self, key: str, full_key: bool = False, chunk_length: int = 1024 * 1024
    ) -> AsyncIterator[bytes]:
        if full_key is False:
            key = self._build_key(key)

        response = await self._client.get_object(Bucket=self._bucket, Key=key)
        # this will ensure the connection is correctly re-used/closed
        body_stream: StreamingBody = response["Body"]
        async for chunk in body_stream.iter_chunks(chunk_length):
            yield chunk

    async def get_object(self, key: str, full_key: bool = False) -> bytes:
        if full_key is False:
            key = self._build_key(key)
        response = await self._client.get_object(Bucket=self._bucket, Key=key)
        # this will ensure the connection is correctly re-used/closed
        async with response["Body"] as stream:
            return await stream.read()

    async def get_object_attributes(self, key: str, full_key: bool = False) -> ObjectAttributes:
        if full_key is False:
            key = self._build_key(key)
        response = await self._client.get_object_attributes(
            Bucket=self._bucket,
            Key=key,
            ObjectAttributes=["LastModified", "ObjectSize"],
        )
        return ObjectAttributes(
            last_modified=response["LastModified"],
            object_size=response["ObjectSize"],
        )

    async def list_objects(self, prefix: str = "") -> List[str]:
        objects = []
        paginator = self._client.get_paginator("list_objects_v2")
        async for result in paginator.paginate(Bucket=self._bucket, Prefix=self._build_key(prefix)):
            objects += [obj["Key"][len(self.key_prefix) :] for obj in result.get("Contents", [])]

        return objects

    async def copy_object(self, origin: str, destination: str) -> None:
        copy_source = {
            "Bucket": self._bucket,
            "Key": self._build_key(origin),
        }
        await self._client.copy_object(Bucket=self._bucket, Key=self._build_key(destination), CopySource=copy_source)

    async def delete_object(self, name):
        key = self._build_key(name)
        await self._client.delete_object(Bucket=self._bucket, Key=key)

    async def create_presigned_url(self, key_name: str, expiration: int = 3600) -> str:
        """Create a temporary public download url for the given `key_name` and `expiration` in seconds.

        Args:
            key_name (str): Full path to the object in the bucket
            expiration (int, optional): Url expiration time in seconds. Defaults to 3600.

        Returns:
            str: Download url to the object with the given expiration and key_name
        """
        return await self._client.generate_presigned_url(
            "get_object",
            Params=dict(
                Bucket=self._bucket,
                Key=key_name,
            ),
            ExpiresIn=expiration,
        )


async def _group_chunks_to_min_part_size(in_iter: AsyncIterator[bytes]) -> AsyncIterator[bytes]:
    current_chunk: bytes = b""
    async for chunk in in_iter:
        current_chunk += chunk
        if len(current_chunk) >= _MIN_PART_SIZE:
            yield current_chunk
            current_chunk = b""

    yield current_chunk



#! /usr/bin/env python

from argparse import ArgumentParser
import re
from typing import Dict, Optional
import hashlib
from pathlib import Path
import asyncio
from ..utils import run_os_cmd

LAMBDA_LAYER_REVISION_KEY = "LambdaLayerRevision"
LOCK_PATH = Path("Pipfile.lock")


async def get_remote_revision(stage: str) -> Optional[str]:
    cmd = f"npx sls info -s {stage} -c lambda_layer.yml --verbose"
    try:
        cmd_out = await run_os_cmd(cmd, echo=False)
    except RuntimeError as e:
        if re.search(r"Stack with id .* does not exist", str(e)):
            return None
        raise e
    out_lines = cmd_out.splitlines()

    revision_line = next(
        (line for line in out_lines if LAMBDA_LAYER_REVISION_KEY in line),
        None,
    )

    if revision_line is None:
        return None

    return revision_line.split(": ")[-1]


def get_local_revision() -> str:
    return hashlib.md5(LOCK_PATH.read_bytes()).hexdigest()


async def deploy(stage: str, local_revision: str) -> None:
    cmd = f"npx sls deploy -c lambda_layer.yml -s {stage}"
    await run_os_cmd(cmd, env={"LOCK_HASH": local_revision})
    print(f"Deployed in {stage}!")


async def deploy_lambda_layer(stage: str, force: bool) -> None:
    local_revision = get_local_revision()
    print(f"{local_revision=}")

    if force:
        print("Forcing deploy")
        return await deploy(stage, local_revision)

    remote_revision = await get_remote_revision(stage)
    print(f"{remote_revision=}")

    if local_revision == remote_revision:
        print("Versions are the same, let's do nothing!")
        return
    print("Versions are different, let's deploy!")
    await deploy(stage, local_revision)


def main():
    parser = ArgumentParser()
    parser.add_argument("stage", type=str, help="Layer stage")
    parser.add_argument("--force", action="store_true")
    args = parser.parse_args()
    asyncio.run(deploy_lambda_layer(args.stage, args.force))



from asyncio import Protocol
from typing import Any, Optional, Dict


class Publisher(Protocol):
    async def publish(self, message: Any, attributes: Optional[Dict[str, str]] = None) -> None:
        """Protocol function"""



from dataclasses import dataclass
from typing import Any, Optional, TypedDict


class ErrorUnit(TypedDict):
    identifier: Optional[str]
    msg: str
    loc: str
    type: str


@dataclass
class ClubbiValidationError(Exception):
    errors: list[ErrorUnit]
    model: str

    def dict(self) -> dict[str, Any]:
        return dict(
            errors=self.errors,
            model=self.model,
        )



from datetime import date, datetime
from typing import Any, Callable, Generator, Union


class BrDate(date):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], date], None, None]:
        yield cls._validate

    @classmethod
    def _validate(cls, value: Any) -> date:
        if isinstance(value, date):
            return value
        if isinstance(value, str):
            return datetime.strptime(value, "%d/%m/%Y").date()
        raise ValueError(f"Unrecognized type {value}")


BrCompatibleDate = Union[BrDate, date]



from typing import Optional

from sqlalchemy.ext.asyncio import AsyncEngine

from clubbi_utils.sqlalchemy.connector import SqlAlchemyConfig, SessionMaker, create_sqlalchemy_session_maker


class PostgresConfig(SqlAlchemyConfig):
    port: int = 5432
    driver_name: str = "postgresql+asyncpg"

    class Config:
        env_prefix = "postgres_"


def create_postgres_session_maker(
    *, engine: Optional[AsyncEngine] = None, config: Optional[PostgresConfig] = None, expire_on_commit: bool = False
) -> SessionMaker:
    if engine is None and config is None:
        config = PostgresConfig()
    return create_sqlalchemy_session_maker(engine=engine, config=config, expire_on_commit=expire_on_commit)



import time
from inspect import iscoroutinefunction
from typing import Callable, TypeVar, cast
from clubbi_utils.json_logging import jlogger


def _log(f: Callable, begin: float, end: float) -> None:
    meta = dict(
        qualname=f.__qualname__,
        module=f.__module__,
    )
    jlogger.info(f.__name__, "time_elapsed", duration=end - begin, meta=meta)


FUNCTION_TYPE = TypeVar("FUNCTION_TYPE", bound=Callable)


def with_timeit(f: FUNCTION_TYPE) -> FUNCTION_TYPE:
    def wrapper(*args, **kwargs):
        begin = time.time()
        r = f(*args, **kwargs)
        end = time.time()
        _log(f, begin, end)
        return r

    async def async_wrapper(*args, **kwargs):
        begin = time.time()
        r = await f(*args, **kwargs)
        end = time.time()
        _log(f, begin, end)
        return r

    if iscoroutinefunction(f):
        return cast(FUNCTION_TYPE, async_wrapper)
    else:
        return cast(FUNCTION_TYPE, wrapper)



from typing import TypeVar, Union


class Undefined(object):
    def __bool__(self):
        return False

    def __eq__(self, other):
        return type(self) == type(other)

    def __copy__(self):
        return self

    def __deepcopy__(self, _):
        return self

    def __hash__(self):
        return hash(None) + 1


UNDEFINED = Undefined()

T = TypeVar("T")

Undefinable = Union[T, Undefined]
